from collections import defaultdict
from os.path import abspath, basename, splitext
from pathlib import Path
from typing import Dict, List

import click
import yaml
from mach import exceptions
from mach.types import (
    CloudOption,
    ComponentAzureConfig,
    ComponentConfig,
    Endpoint,
    MachConfig,
    SentryDsn,
    ServicePlan,
    Site,
    SiteAzureSettings,
)
from mach.validate import validate_config
from marshmallow.exceptions import ValidationError


def parse_configs(
    files: List[str], output_path: str = None, *, ignore_version=True
) -> List[MachConfig]:
    """Parse and validate configurations."""
    valid_configs = []
    for file in files:
        config = parse_config_from_file(file)
        config.file = file
        click.echo(f"Parsed {file} into config")
        validate_config(config, ignore_version=ignore_version)

        if output_path:
            full_output_path = Path(f"{output_path}/{splitext(basename(file))[0]}")
            full_output_path.mkdir(exist_ok=True, parents=True)
            config.output_path = str(full_output_path)

        valid_configs.append(config)
    return valid_configs


def parse_config_from_file(file: str) -> MachConfig:
    """Parse file into MachConfig object."""
    click.echo(f"Parsing {file}...")
    with open(file, "r") as fh:
        dictionary_config = yaml.full_load(fh)

    try:
        config = MachConfig.schema(infer_missing=True).load(dictionary_config)  # type: ignore
    except KeyError as e:
        # Most probably a missing value in the configuration.
        # dataclasses_json doesn't really give a proper Exception for this.
        # TODO: See if we can improve this / make it more robust. Either by improving
        # dataclassess_json (with a PR) or by extending it (if possible)
        raise exceptions.ParseError(f"Required attribute {e} missing") from e
    except ValidationError as e:
        # TODO: We don't have any path here, so not the best of error messages
        raise exceptions.ParseError(
            "Configuration file could not be validated", details=e.normalized_messages()
        ) from e

    return parse_config(config)


def parse_config(config: MachConfig) -> MachConfig:
    parse_global_config(config)
    resolve_component_definitions(config)
    resolve_site_configs(config)
    return config


def parse_global_config(config: MachConfig):
    if config.general_config.cloud == CloudOption.AZURE:
        if "default" not in config.general_config.azure.service_plans:
            config.general_config.azure.service_plans["default"] = ServicePlan(
                kind="FunctionApp", tier="Dynamic", size="Y1"
            )


# flake8: noqa: C901
def resolve_site_configs(config: MachConfig):
    """Use and merge site-specific configurations with general config."""
    for site in config.sites:
        if config.general_config.cloud == CloudOption.AZURE:
            if site.azure:
                site.azure.merge(config.general_config.azure)
            else:
                site.azure = SiteAzureSettings.from_config(config.general_config.azure)

            if site.azure.resource_group:
                click.echo(
                    click.style(
                        (
                            f"WARNING: resource_group on {site.identifier} "
                            f"is used ({site.azure.resource_group}). "
                        ),
                        fg="red",
                        bold=True,
                    )
                )
                click.echo(
                    click.style(
                        (
                            "   Make sure it wasn't managed by MACH before otherwise "
                            "the resource group will get deleted."
                        ),
                        fg="red",
                    )
                )

        # Merge Contentful settings
        if config.general_config.contentful:
            for site in config.sites:
                if site.contentful:
                    site.contentful.merge(config.general_config.contentful)

        # Merge Amplience settings
        if config.general_config.amplience:
            for site in config.sites:
                if site.amplience:
                    site.amplience.merge(config.general_config.amplience)

        if config.general_config.sentry:
            if not site.sentry:
                site.sentry = SentryDsn.from_config(config.general_config.sentry)
            else:
                site.sentry.merge(config.general_config.sentry)

    resolve_site_components(config)

    for site in config.sites:
        resolve_endpoint_components(site)
        if site.azure:
            resolve_used_service_plans(site)


def resolve_used_service_plans(site: Site):
    "Azure-specific method to find out which service plans are actually being used by components"
    used = [
        c.azure.service_plan
        for c in site.components
        if c.azure and c.azure.service_plan
    ]
    site.azure.service_plans = {
        key: sp for key, sp in site.azure.service_plans.items() if key in used
    }


def resolve_endpoint_components(site: Site):
    endpoint_components = defaultdict(list)
    for c in site.components:
        for endpoint in c.endpoints.values():
            endpoint_components[endpoint].append(c)

    site_endpoint_keys = {e.key for e in site.endpoints}
    # If one of the components has a 'default' endpoint defined,
    # we'll include it to our site endpoints.
    # A 'default' endpoint is one without a custom domain, so no further
    # Route53 or DNS zone settings required.
    if "default" in endpoint_components and not "default" in site_endpoint_keys:
        click.echo(
            click.style(
                (
                    "WARNING: 'default' endpoint used but not defined in the site endpoints.\n"
                    "MACH will create a default endpoint without any custom domain attached to it.\n"
                    "More info: https://docs.machcomposer.io/reference/syntax/sites.html#endpoints"
                ),
                fg="yellow",
            )
        )
        site.endpoints.append(
            Endpoint(
                url="",
                key="default",
            )
        )

    for endpoint in site.endpoints:
        endpoint.components = endpoint_components.get(endpoint.key, [])


def resolve_site_components(config: MachConfig) -> MachConfig:
    """If no component info is specified, use global component settings."""
    component_info: Dict[str, ComponentConfig] = {
        component.name: component for component in config.components
    }
    for site in config.sites:
        if not site.components:
            continue

        for component in site.components:
            try:
                info = component_info[component.name]
            except KeyError:
                raise exceptions.ParseError(
                    f"Component {component.name} does not exist in global components."
                )
            component.definition = info

            if site.sentry:
                if not component.sentry:
                    component.sentry = site.sentry
                else:
                    component.sentry.merge(site.sentry)

            if site.azure:
                if not component.azure:
                    component.azure = info.azure
                else:
                    component.azure.merge(info.azure)

    return config


def resolve_component_definitions(config: MachConfig) -> MachConfig:
    for comp in config.components:
        # Terraform needs absolute paths to modules
        if comp.source.startswith("."):
            comp.source = abspath(comp.source)

        if not comp.integrations:
            # If no integrations are given, set the Cloud integrations as default
            if config.general_config.cloud == CloudOption.AWS:
                comp.integrations = ["aws"]
            elif config.general_config.cloud == CloudOption.AZURE:
                comp.integrations = ["azure"]

        if config.general_config.cloud == CloudOption.AZURE:
            if not comp.azure:
                comp.azure = ComponentAzureConfig()
            if not comp.azure.short_name:
                comp.azure.short_name = comp.name



from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, List, Optional

from dataclasses_json import config, dataclass_json
from dataclasses_jsonschema import JsonSchemaMixin

from . import fields
from .components import ComponentConfig
from .general_config import GlobalConfig
from .mach import MachComposerConfig
from .sites import Site

__all__ = ["MachConfig"]


@dataclass_json
@dataclass
class MachConfig(JsonSchemaMixin):
    """Main MACH configuration object."""

    mach_composer: MachComposerConfig
    general_config: GlobalConfig = field(metadata=config(field_name="global"))
    sites: List[Site]
    components: List[ComponentConfig] = fields.list_()

    # Not used during a mach update or apply, but MACH
    # must be able to accept this attribute when parsing
    # encrypted configurations
    sops: Optional[Any] = fields.none()

    # Items that are not used in the configuration itself by set by the parser
    output_path: str = "deployments"
    file: Optional[str] = fields.none()

    @property
    def deployment_path(self) -> Path:
        return Path(self.output_path)

    def get_component(self, name: str) -> Optional[ComponentConfig]:
        for comp in self.components:
            if comp.name == name:
                return comp
        return None



import os
import subprocess
from dataclasses import dataclass
from typing import List

import click
from mach import exceptions

PRETTY_FMT = {
    "commit": "%H",
    "author": "%aN <%aE>",
    "date": "%ad",
    "message": "%s",
}

PRETTY_FMT_STR = "format:" + "|".join([fmt for fmt in PRETTY_FMT.values()])


@dataclass
class Commit:
    id: str
    msg: str


def commit(message: str):
    result = _run(["git", "status", "--short"])
    if not result:
        click.echo("No changes detected, won't commit anything")
        return

    _run(["git", "commit", "-m", message])


def add(file: str):
    _run(["git", "add", file])


def ensure_local(repo: str, dest: str, *, reference: str = ""):
    """Ensure the repository is present on the given dest."""
    if os.path.exists(dest):
        _run(["git", "pull"], cwd=dest)
    else:
        clone(repo, dest)

    if reference:
        _run(["git", "reset", "--hard", reference], cwd=dest)


def clone(repo: str, dest: str):
    _run(["git", "clone", repo, dest])


def history(dir: str, from_ref: str, *, branch: str = "") -> List[Commit]:
    if branch:
        _run(["git", "checkout", branch], cwd=dir)

    cmd = ["git", "log", f"--pretty={PRETTY_FMT_STR}"]
    if from_ref:
        cmd.append(f"{from_ref}..{branch or ''}")

    lines = _run(cmd, cwd=dir).decode("utf-8").splitlines()
    commits = []
    for line in lines:
        commit_id, author, date, message = line.split("|")
        commits.append(
            Commit(id=_clean_commit_id(commit_id), msg=_clean_commit_msg(message))
        )

    return commits


def _clean_commit_msg(msg: str) -> str:
    return msg


def _clean_commit_id(commit_id: str) -> str:
    """Get the correct commit ID for this commit.

    It will trim the short_id since mach and the components are using a
    different commit id format (7 chars long).
    """
    return commit_id[:7]


def _run(cmd: List, *args, **kwargs) -> str:
    kwargs["stderr"] = subprocess.DEVNULL

    try:
        return subprocess.check_output(cmd, *args, **kwargs)
    except subprocess.CalledProcessError as e:
        raise exceptions.MachError(f"Could not perform command: {e}")



import os
import re
import typing

import click
from mach import cache, exceptions, git
from mach.types import ComponentConfig, MachConfig

NAME_RE = re.compile(r".* name: [\"']?(.*)[\"']?")
VERSION_RE = re.compile(r"(\s*version: )([\"']?.*[\"']?)")

ROOT_BLOCK_START = re.compile(r"^\w+")


Updates = typing.List[typing.Tuple[ComponentConfig, str]]


def update_config_component(  # noqa: C901
    config: MachConfig,
    component_name: str,
    new_version: str,
):
    component = config.get_component(component_name)
    if not component:
        raise exceptions.MachError(f"Could not find component {component_name}")

    if component.version == new_version:
        click.echo(f"Component {component_name} is already on version {new_version}.")

    click.echo(f"Updating {component_name} to version {new_version}...")

    FileUpdater.apply_updates(config.file, [(component, new_version)])


def update_config_components(  # noqa: C901
    config: MachConfig,
    *,
    check_only=False,
    verbose=False,
):
    """
    Update a given MACH configuration file.

    :param config: The MACH configuration to update components for
    :param check_only: Only check for updates; don't update the file
    :param verbose: Enable verbose output
    """
    intro_msg = f"Checking updates for components in {config.file}"
    print(intro_msg)
    print("-" * len(intro_msg))

    updates: Updates = _fetch_changes(config)

    if not check_only:
        FileUpdater.apply_updates(config.file, updates)


def _fetch_changes(config: MachConfig) -> Updates:
    cache_dir = cache.cache_dir_for(config)
    updates: Updates = []

    for component in config.components:
        click.echo(f"Updates for {component.name}...")

        match = re.match(r"^git::(.*)", component.source)
        if not match:
            click.echo(
                f"  Cannot check {component.name} component since it doesn't have a Git source defined"  # noqa
            )
            continue

        component_dir = os.path.join(cache_dir, component.name)
        repo = match.group(1)
        match = re.match(r"(.*\/.*)(\/\/.*)$", repo)
        if match:
            repo = match.group(1)
        git.ensure_local(repo, component_dir)

        commits = git.history(component_dir, component.version, branch=component.branch)
        if not commits:
            click.echo("  No updates\n")
            continue

        for commit in commits:
            print(f"  {commit.id}: {commit.msg}")

        click.echo("")

        updates.append((component, commits[0].id))

    return updates


class FileUpdater:
    """Updater which update component version in-place.

    We'll use a very basic search-and-replace based on regular expressions
    instead of the yaml parser to not mess with any formatting.
    """

    @classmethod
    def apply_updates(cls, file: str, updates: Updates):
        """Apply given updates to the file."""
        instance = cls()
        instance._apply_updates(file, updates)

    def _apply_updates(self, file: str, updates: Updates):
        self.in_components = False
        self.current_component: typing.Optional[ComponentConfig] = None
        self.updates = {component.name: version for component, version in updates}
        self.component_map = {component.name: component for component, _ in updates}

        with open(file) as f:
            lines = f.readlines()

        newlines = []
        for line in lines:
            if line.startswith("components:"):
                self.in_components = True
            elif ROOT_BLOCK_START.match(line):
                self.in_components = False
            elif self.in_components:
                line = self.process_component_line(line)
            newlines.append(line)

        with open(file, mode="w") as f:
            for line in newlines:
                f.write(line)

    def process_component_line(self, line: str):
        name_match = NAME_RE.match(line)

        if name_match:
            component_name = name_match.group(1)

            try:
                self.current_component = self.component_map[component_name]
            except KeyError:
                self.current_component = None

            return line

        if not self.current_component:
            return line

        match = VERSION_RE.match(line)
        if not match:
            return line

        assert self.current_component.version in match.group(2)

        try:
            new_version = self.updates[self.current_component.name]
        except KeyError:
            return line

        if new_version.isdigit():
            new_version = f'"{new_version}"'

        return VERSION_RE.sub(rf"\g<1>{new_version}", line)



import glob
import subprocess
import sys
from functools import update_wrapper
from typing import List, Optional

import click
from mach import bootstrap as _bootstrap
from mach import git, parse, updater
from mach.build import build_packages
from mach.exceptions import MachError
from mach.terraform import apply_terraform, generate_terraform, plan_terraform


@click.group()
def mach():
    pass


def terraform_command(f):
    @click.option(
        "-f",
        "--file",
        default=None,
        help="YAML file to parse. If not set parse all *.yml files.",
    )
    @click.option(
        "-s",
        "--site",
        default=None,
        help="Site to parse. If not set parse all sites.",
    )
    @click.option(
        "--ignore-version",
        is_flag=True,
        default=False,
        help="Skip MACH composer version check",
    )
    @click.option(
        "--output-path",
        default="deployments",
        help="Output path, defaults to `cwd`/deployments.",
    )
    def new_func(file, site, output_path: str, ignore_version: bool, **kwargs):
        files = get_input_files(file)

        try:
            configs = parse.parse_configs(
                files, output_path, ignore_version=ignore_version
            )
        except MachError as e:
            raise click.ClickException(str(e)) from e

        try:
            result = f(
                file=file,
                site=site,
                configs=configs,
                **kwargs,
            )
        except subprocess.CalledProcessError as e:
            click.echo("Failed to run")
            sys.exit(e.returncode)
        else:
            click.echo("Done 👍")
            return result

    return update_wrapper(new_func, f)


@mach.command()
@terraform_command
def generate(file, site, configs, *args, **kwargs):
    """Generate the Terraform files."""
    for config in configs:
        generate_terraform(config, site=site)


@mach.command()
@click.option(
    "--with-sp-login",
    is_flag=True,
    default=False,
    help="If az login with service principal environment variables "
    "(ARM_CLIENT_ID, ARM_CLIENT_SECRET, ARM_TENANT_ID) should be done.",
)
@click.option(
    "-c",
    "--component",
    multiple=True,
    default=False,
    help="",
)
@click.option(
    "--reuse",
    default=False,
    is_flag=True,
    help="Supress a terraform init for improved speed (not recommended for production usage)",
)
@terraform_command
def plan(file, site, configs, with_sp_login, component, reuse, *args, **kwargs):
    """Output the deploy plan."""
    for config in configs:
        generate_terraform(config, site=site)
        plan_terraform(
            config,
            site=site,
            components=component,
            with_sp_login=with_sp_login,
            reuse=reuse,
        )


@mach.command()
@click.option(
    "--with-sp-login",
    is_flag=True,
    default=False,
    help="If az login with service principal environment variables "
    "(ARM_CLIENT_ID, ARM_CLIENT_SECRET, ARM_TENANT_ID) should be done.",
)
@click.option(
    "--auto-approve",
    is_flag=True,
    default=False,
    help="",
)
@click.option(
    "-c",
    "--component",
    multiple=True,
    default=False,
    help="",
)
@click.option(
    "--reuse",
    is_flag=True,
    default=False,
    help="Supress a terraform init for improved speed (not recommended for production usage)",
)
@terraform_command
def apply(
    file,
    site,
    configs,
    with_sp_login,
    auto_approve,
    component,
    reuse,
    *args,
    **kwargs,
):
    """Apply the configuration."""
    for config in configs:
        build_packages(config, restrict_components=component)
        generate_terraform(config, site=site)
        apply_terraform(
            config,
            site=site,
            components=component,
            with_sp_login=with_sp_login,
            auto_approve=auto_approve,
            reuse=reuse,
        )


@mach.command()
@click.option(
    "-f",
    "--file",
    default=None,
    help="YAML file to update. If not set update all *.yml files.",
)
@click.option("-v", "--verbose", is_flag=True, default=False, help="Verbose output.")
@click.option(
    "--check",
    default=False,
    is_flag=True,
    help="Only checks for updates, doesnt change files.",
)
@click.option(
    "-c",
    "--commit",
    default=False,
    is_flag=True,
    help="Automatically commits the change.",
)
@click.argument("component", required=False)
@click.argument("version", required=False)
def update(
    file: str,
    check: bool,
    verbose: bool,
    component: str,
    version: str,
    commit: bool,
):
    """Update all (or a given) component.

    When no component and version is given, it will check the git repositories for any updates.
    This command can also be used to manually update a single component by specifying a component
    and version.
    """
    if check and commit:
        raise click.ClickException(
            "check_only is not possible when create_commit is enabled."
        )

    if component and not version:
        raise click.ClickException(
            f"When specifying a component ({component}) you should specify a version as well"
        )

    files = get_input_files(file)
    configs = parse.parse_configs(files)
    try:
        for config in configs:
            if component and version:
                updater.update_config_component(config, component, version)
            else:
                updater.update_config_components(
                    config, verbose=verbose, check_only=check
                )
            if commit:
                git.add(config.file)

        if commit:
            if component:
                commit_msg = f"Updated {component} component"
            else:
                commit_msg = "Updated components"

            git.commit(commit_msg)

    except MachError as e:
        raise click.ClickException(str(e)) from e


@mach.command()
@click.option(
    "-f",
    "--file",
    default=None,
    help="YAML file to read. If not set read all *.yml files.",
)
def components(file: str):
    """List all components."""
    files = get_input_files(file)
    configs = parse.parse_configs(files)
    for config in configs:
        click.echo(f"{config.file}:")
        for component in config.components:
            click.echo(f" - {component.name}")
            click.echo(f"   version: {component.version}")

        click.echo("")


@mach.command()
@click.option(
    "-f",
    "--file",
    default=None,
    help="YAML file to read. If not set read all *.yml files.",
)
def sites(file: str):
    """List all sites."""
    files = get_input_files(file)
    configs = parse.parse_configs(files)
    for config in configs:
        click.echo(f"{config.file}:")
        for site in config.sites:
            click.echo(f" - {site.identifier}")
            click.echo("   components:")
            for component in site.components:
                click.echo(f"     {component.name}")

        click.echo("")


@mach.command()
@click.option(
    "-o",
    "--output",
    help="Output file or directory.",
)
@click.option(
    "-c",
    "--cookiecutter",
    default="",
    help="cookiecutter repository to generate from.",
)
@click.argument("type_", required=True, type=click.Choice(["config", "component"]))
def bootstrap(output: str, type_: str, cookiecutter: str):
    """Bootstraps a configuration or component."""
    if type_ == "config":
        _bootstrap.create_configuration(output or "main.yml")
    if type_ == "component":
        _bootstrap.create_component(output, cookiecutter)


def get_input_files(file: Optional[str]) -> List[str]:
    """Determine input files. If file is not specified use all *.yml files."""
    if file:
        files = [file]
    else:
        files = glob.glob("./*.yml")
    if not files:
        click.echo("No .yml files found")
        sys.exit(1)
    return files
